@@ -110,7 +110,7 @@ contrast colour with RGB as background and as foreground."
 ;;;;; Commands
 
 (cl-defun ement-create-room
-    (session &key name alias topic invite direct-p
+    (session &key name alias topic invite direct-p creation-content
              (then (lambda (data)
                      (message "Created new room: %s" (alist-get 'room_id data))))
              (visibility 'private))
@@ -118,7 +118,8 @@ contrast colour with RGB as background and as foreground."
 Then call function THEN with response data.  Optional string
 arguments are NAME, ALIAS, and TOPIC.  INVITE may be a list of
 user IDs to invite.  If DIRECT-P, set the \"is_direct\" flag in
-the request."
+the request.  CREATION-CONTENT may be an alist of extra keys to
+include with the request (see Matrix spec)."
   ;; TODO: Document other arguments.
   ;; SPEC: 10.1.1.
   (declare (indent defun))
@@ -126,7 +127,7 @@ the request."
 		     :name (read-string "New room name: ")
 		     :alias (read-string "New room alias (e.g. \"foo\" for \"#foo:matrix.org\"): ")
 		     :topic (read-string "New room topic: ")
-		     :visibility (completing-read "New room type: " '(private public))))
+		     :visibility (completing-read "New room visibility: " '(private public))))
   (cl-labels ((given-p
 	       (var) (and var (not (string-empty-p var)))))
     (pcase-let* ((endpoint "createRoom")
@@ -141,17 +142,38 @@ the request."
 			 (when invite
 			   (push (cons "invite" invite) it))
 			 (when direct-p
-			   (push (cons "is_direct" t) it)))))
+			   (push (cons "is_direct" t) it))
+                         (when creation-content
+                           (push (cons "creation_content" creation-content) it)))))
       (ement-api session endpoint :method 'post :data (json-encode data)
         :then then))))
 
+(cl-defun ement-create-space
+    (session &key name alias topic
+             (then (lambda (data)
+                     (message "Created new space: %s" (alist-get 'room_id data))))
+             (visibility 'private))
+  "Create new space on SESSION.
+Then call function THEN with response data.  Optional string
+arguments are NAME, ALIAS, and TOPIC."
+  (declare (indent defun))
+  (interactive (list (ement-complete-session)
+		     :name (read-string "New space name: ")
+		     :alias (read-string "New space alias (e.g. \"foo\" for \"#foo:matrix.org\"): ")
+		     :topic (read-string "New space topic: ")
+		     :visibility (completing-read "New space visibility: " '(private public))))
+  (ement-create-room session :name name :alias alias :topic topic :visibility visibility
+    :creation-content (ement-alist "type" "m.space") :then then))
+
 (defun ement-room-leave (room session &optional force-p)
   "Leave ROOM on SESSION.
 If FORCE-P, leave without prompting.  ROOM may be an `ement-room'
 struct, or a room ID or alias string."
   ;; TODO: Rename `room' argument to `room-or-id'.
-  (interactive (ement-complete-room :session (ement-complete-session)))
-  (cl-assert room) (cl-assert session)
+  (interactive
+   (ement-with-room-and-session
+     :prompt-form (ement-complete-room :prompt "Leave room: ")
+     (list ement-room ement-session)))
   (cl-etypecase room
     (ement-room)
     (string (setf room (ement-afirst (or (equal room (ement-room-canonical-alias it))
@@ -175,7 +197,7 @@ struct, or a room ID or alias string."
                                          (kill-buffer buffer))))))
                     (setf (symbol-function leave-fn-symbol) leave-fn)
                     (add-hook 'ement-sync-callback-hook leave-fn-symbol)))
-                (message "Left room: %s" (ement--format-room room)))
+                (ement-message "Left room: %s" (ement--format-room room)))
         :else (lambda (plz-error)
                 (pcase-let* (((cl-struct plz-error response) plz-error)
                              ((cl-struct plz-response status body) response)
@@ -189,8 +211,10 @@ struct, or a room ID or alias string."
   "Forget ROOM on SESSION.
 If FORCE-P (interactively, with prefix), prompt to leave the room
 when necessary, and forget the room without prompting."
-  (interactive (pcase-let ((`(,room ,session) (ement-complete-room)))
-                 (list room session current-prefix-arg)))
+  (interactive
+   (ement-with-room-and-session
+     :prompt-form (ement-complete-room :prompt "Forget room: ")
+     (list ement-room ement-session current-prefix-arg)))
   (pcase-let* (((cl-struct ement-room id display-name status) room)
                (endpoint (format "rooms/%s/forget" (url-hexify-string id))))
     (pcase status
@@ -219,7 +243,7 @@ when necessary, and forget the room without prompting."
                           (setf (ement-session-rooms session)
                                 (cl-remove room (ement-session-rooms session)))
                           ;; TODO: Indicate forgotten in footer in room buffer.
-                          (message "Room \"%s\" (%s) forgotten." display-name id))))))))
+                          (ement-message "Forgot room: %s." (ement--format-room room)))))))))
 
 (defun ement-ignore-user (user-id session &optional unignore-p)
   "Ignore USER-ID on SESSION.
@@ -243,13 +267,13 @@ If UNIGNORE-P (interactively, with prefix), un-ignore USER."
               (message "Ement: User %s %s." user-id (if unignore-p "unignored" "ignored"))))))
 
 (defun ement-invite-user (user-id room session)
-  "Invite USER-ID to ROOM on SESSION."
+  "Invite USER-ID to ROOM on SESSION.
+Interactively, with prefix, prompt for room and session,
+otherwise use current room."
   ;; SPEC: 10.4.2.1.
   (interactive
-   (let* ((session (ement-complete-session))
-          (user-id (ement-complete-user-id))
-          (room (car (ement-complete-room :session session))))
-     (list user-id room session)))
+   (ement-with-room-and-session
+     (list (ement-complete-user-id) ement-room ement-session)))
   (pcase-let* ((endpoint (format "rooms/%s/invite"
                                  (url-hexify-string (ement-room-id room))))
                (data (ement-alist "user_id" user-id) ))
@@ -262,9 +286,13 @@ If UNIGNORE-P (interactively, with prefix), un-ignore USER."
 
 (defun ement-list-members (room session bufferp)
   "Show members of ROOM on SESSION.
-If BUFFERP (interactively, with prefix), or if there are many
-members, show in a new buffer; otherwise show in echo area."
-  (interactive (list ement-room ement-session current-prefix-arg))
+Interactively, with prefix, prompt for room and session,
+otherwise use current room.  If BUFFERP (interactively, with
+prefix), or if there are many members, show in a new buffer;
+otherwise show in echo area."
+  (interactive
+   (ement-with-room-and-session
+     (list ement-room ement-session current-prefix-arg)))
   (pcase-let* (((cl-struct ement-room members (local (map fetched-members-p))) room)
                (list-members
                 (lambda (&optional _)
@@ -341,29 +369,38 @@ new one automatically if necessary."
                 (message "Room \"%s\" created for user %s.  Sending message..."
 	                 room-id user-id))))))
 
-(defun ement-tag-room (tag room session &optional delete)
-  "Add TAG to ROOM on SESSION.
-If DELETE (interactively, with prefix), delete it."
+(defun ement-tag-room (tag room session)
+  "Toggle TAG for ROOM on SESSION."
   (interactive
-   (pcase-let* ((`(,room ,session) (or (when (bound-and-true-p ement-room)
-                                         (list ement-room ement-session))
-                                       (ement-complete-room)))
-                (prompt (if current-prefix-arg "Delete tag: " "Add tag: "))
-                (default-tags (ement-alist "Favourite" "m.favourite"
-                                           "Low-priority" "m.lowpriority"))
-                (input (completing-read prompt default-tags))
-                (tag (alist-get input default-tags (concat "u." input) nil #'string=)))
-     (list tag room session current-prefix-arg)))
+   (ement-with-room-and-session
+     (let* ((prompt (format "Toggle tag (%s): " (ement--format-room ement-room)))
+            (default-tags
+              (ement-alist (propertize "Favourite"
+                                       'face (when (ement--room-tagged-p "m.favourite" ement-room)
+                                               'transient-value))
+                           "m.favourite"
+                           (propertize "Low-priority"
+                                       'face (when (ement--room-tagged-p "m.lowpriority" ement-room)
+                                               'transient-value))
+                           "m.lowpriority"))
+            (input (completing-read prompt default-tags))
+            (tag (alist-get input default-tags (concat "u." input) nil #'string=)))
+       (list tag ement-room ement-session))))
   (pcase-let* (((cl-struct ement-session user) session)
                ((cl-struct ement-user (id user-id)) user)
                ((cl-struct ement-room (id room-id)) room)
                (endpoint (format "user/%s/rooms/%s/tags/%s"
                                  (url-hexify-string user-id) (url-hexify-string room-id) (url-hexify-string tag)))
-               (method (if delete 'delete 'put)))
+               (method (if (ement--room-tagged-p tag room) 'delete 'put)))
     ;; TODO: "order".
-    (ement-api session endpoint :version "v3" :method method :data (unless delete "{}")
-      :then (lambda (data)
-              (ement-debug "Changed tag on room" method tag data room)))))
+    ;; FIXME: Removing a tag on a left room doesn't seem to work (e.g. to unfavorite a room after leaving it, but not forgetting it).
+    (ement-api session endpoint :version "v3" :method method :data (pcase method ('put "{}"))
+      :then (lambda (_)
+              (ement-message "%s tag %S on %s"
+                             (pcase method
+                               ('delete "Removed")
+                               ('put "Added"))
+                             tag (ement--format-room room)) ))))
 
 (defun ement-set-display-name (display-name session)
   "Set DISPLAY-NAME for user on SESSION.
@@ -384,16 +421,16 @@ Sets global displayname."
 
 (defun ement-room-set-display-name (display-name room session)
   "Set DISPLAY-NAME for user in ROOM on SESSION.
-Sets the name only in ROOM, not globally."
+Interactively, with prefix, prompt for room and session,
+otherwise use current room.  Sets the name only in ROOM, not
+globally."
   (interactive
-   (pcase-let* ((`(,room ,session) (or (when (bound-and-true-p ement-room)
-                                         (list ement-room ement-session))
-                                       (ement-complete-room)))
-                (prompt (format "Set display-name in %S to: "
-                                (ement--format-room room)))
-                (display-name (read-string prompt nil nil
-                                           (ement-user-displayname (ement-session-user session)))))
-     (list display-name room session)))
+   (ement-with-room-and-session
+     (let* ((prompt (format "Set display-name in %S to: "
+                            (ement--format-room ement-room)))
+            (display-name (read-string prompt nil nil
+                                       (ement-user-displayname (ement-session-user ement-session)))))
+       (list display-name ement-room ement-session))))
   ;; NOTE: This does not seem to be documented in the spec, so we imitate the
   ;; "/myroomnick" command in SlashCommands.tsx from matrix-react-sdk.
   (pcase-let* (((cl-struct ement-room state) room)
@@ -424,9 +461,10 @@ Sets the name only in ROOM, not globally."
   "Ement-Describe-Room" "Major mode for `ement-describe-room' buffers.")
 
 (defun ement-describe-room (room session)
-  "Describe ROOM on SESSION."
-  (interactive (pcase-let ((`(,room ,session) (ement-complete-room :session ement-session)))
-                 (list room session)))
+  "Describe ROOM on SESSION.
+Interactively, with prefix, prompt for room and session,
+otherwise use current room."
+  (interactive (ement-with-room-and-session (list ement-room ement-session)))
   (cl-labels ((heading (string)
                        (propertize (or string "") 'face 'font-lock-builtin-face))
               (id (string)
@@ -556,8 +594,9 @@ Returns one of nil (meaning default rules are used), `all-loud',
 
 (defun ement-room-set-notification-state (state room session)
   "Set notification STATE for ROOM on SESSION.
-STATE may be nil to set the rules to default, `all',
-`mentions-and-keywords', or `none'."
+Interactively, with prefix, prompt for room and session,
+otherwise use current room.  STATE may be nil to set the rules to
+default, `all', `mentions-and-keywords', or `none'."
   ;; This merely attempts to reproduce the behavior of Element's simple notification
   ;; options.  It does not attempt to offer all of the features defined in the spec.  And,
   ;; yes, it is rather awkward, having to sometimes* make multiple requests of different
@@ -574,17 +613,15 @@ STATE may be nil to set the rules to default, `all',
 
   ;; TODO: Support `all-loud' ("all_messages_loud").
   (interactive
-   (pcase-let* ((`(,room ,session) (or (when (bound-and-true-p ement-room)
-                                         (list ement-room ement-session))
-                                       (ement-complete-room)))
-                (prompt (format "Set notification rules for %s: " (ement--format-room room)))
-                (available-states (ement-alist "Default" nil
-                                               "All messages" 'all
-                                               "Mentions and keywords" 'mentions-and-keywords
-                                               "None" 'none))
-                (selected-rule (completing-read prompt (mapcar #'car available-states) nil t))
-                (state (alist-get selected-rule available-states nil nil #'equal)))
-     (list state room session)))
+   (ement-with-room-and-session
+     (let* ((prompt (format "Set notification rules for %s: " (ement--format-room ement-room)))
+            (available-states (ement-alist "Default" nil
+                                           "All messages" 'all
+                                           "Mentions and keywords" 'mentions-and-keywords
+                                           "None" 'none))
+            (selected-rule (completing-read prompt (mapcar #'car available-states) nil t))
+            (state (alist-get selected-rule available-states nil nil #'equal)))
+       (list state ement-room ement-session))))
   (cl-labels ((set-rule (kind rule queue message-fn)
                         (pcase-let* (((cl-struct ement-room (id room-id)) room)
                                      (rule-id (url-hexify-string room-id))
@@ -721,12 +758,14 @@ unseen user IDs to be input as well."
 	  selected-user))))
 
 (cl-defun ement-put-account-data
-    (session type data &key
+    (session type data &key room
              (then (lambda (received-data)
                      ;; Handle echoed-back account data event (the spec does not explain this,
                      ;; but see <https://github.com/matrix-org/matrix-react-sdk/blob/675b4271e9c6e33be354a93fcd7807253bd27fcd/src/settings/handlers/AccountSettingsHandler.ts#L150>).
                      ;; FIXME: Make session account-data a map instead of a list of events.
-                     (push received-data (ement-session-account-data session))
+                     (if room
+                         (push received-data (ement-room-account-data room))
+                       (push received-data (ement-session-account-data session)))
 
                      ;; NOTE: Commenting out this ement-debug form because a bug in Emacs
                      ;; causes this long string to be interpreted as the function's
@@ -736,10 +775,12 @@ unseen user IDs to be input as well."
                      ;;              (ement-user-id (ement-session-user session)) (json-encode data) received-data)
                      )))
   "Put account data of TYPE with DATA on SESSION.
-Also handle the echoed-back event."
+If ROOM, put it on that room's account data.  Also handle the
+echoed-back event."
   (declare (indent defun))
   (pcase-let* (((cl-struct ement-session (user (cl-struct ement-user (id user-id)))) session)
-               (endpoint (format "user/%s/account_data/%s" (url-hexify-string user-id) type)))
+               (room-part (if room (format "/rooms/%s" (ement-room-id room)) ""))
+               (endpoint (format "user/%s%s/account_data/%s" (url-hexify-string user-id) room-part type)))
     (ement-api session endpoint :method 'put :data (json-encode data)
       :then then)))
 
@@ -768,10 +809,56 @@ USER is an `ement-user' struct."
 ;; These functions aren't expected to be called by code in other packages (but if that
 ;; were necessary, they could be renamed accordingly).
 
+;; (defun ement--room-routing (room)
+;;   "Return a list of servers to route to ROOM through."
+;;   ;; See <https://spec.matrix.org/v1.2/appendices/#routing>.
+;;   ;; FIXME: Ensure highest power level user is at least level 50.
+;;   ;; FIXME: Ignore servers blocked due to server ACLs.
+;;   ;; FIXME: Ignore servers which are IP addresses.
+;;   (cl-labels ((most-powerful-user-in
+;;                (room))
+;;               (servers-by-population-in
+;;                (room))
+;;               (server-of (user)))
+;;     (let (first-server-by-power-level)
+;;       (delete-dups
+;;        (remq nil
+;;              (list
+;;               ;; 1.
+;;               (or (when-let ((user (most-powerful-user-in room)))
+;;                     (setf first-server-by-power-level t)
+;;                     (server-of user))
+;;                   (car (servers-by-population-in room)))
+;;               ;; 2.
+;;               (if first-server-by-power-level
+;;                   (car (servers-by-population-in room))
+;;                 (cl-second (servers-by-population-in room)))
+;;               ;; 3.
+;;               (cl-third (servers-by-population-in room))))))))
+
 (defun ement--room-space-p (room)
   "Return non-nil if ROOM is a space."
   (equal "m.space" (ement-room-type room)))
 
+(defun ement--room-in-space-p (room space)
+  "Return non-nil if ROOM is in SPACE on SESSION."
+  ;; We could use `ement---room-spaces', but since that returns rooms by looking them up
+  ;; by ID in the session's rooms list, this is more efficient.
+  (pcase-let* (((cl-struct ement-room (id parent-id) (local (map children))) space)
+               ((cl-struct ement-room (id child-id) (local (map parents))) room))
+    (or (member parent-id parents)
+        (member child-id children))))
+
+(defun ement--room-spaces (room session)
+  "Return list of ROOM's parent spaces on SESSION."
+  ;; NOTE: This only looks in the room's parents list; it doesn't look in every space's children
+  ;; list.  This should be good enough, assuming we add to the lists correctly elsewhere.
+  (pcase-let* (((cl-struct ement-session rooms) session)
+               ((cl-struct ement-room (local (map parents))) room))
+    (cl-remove-if-not (lambda (session-room-id)
+                        (member session-room-id parents))
+                      rooms :key #'ement-room-id)))
+
 (cl-defun ement--prism-color (string &key (contrast-with (face-background 'default nil 'default)))
   "Return a computed color for STRING.
 The color is adjusted to have sufficient contrast with the color
@@ -984,7 +1071,7 @@ suggested room."
                                  (when-let ((suggestion (ement--room-at-point)))
                                    (when (or (not predicate)
                                              (funcall predicate suggestion))
-                                     suggestion))))))
+                                     (ement--format-room suggestion 'topic)))))))
     (alist-get selected-name name-to-room-session nil nil #'string=)))
 
 (cl-defun ement-send-message (room session
@@ -1015,16 +1102,18 @@ e.g. `ement-room-send-org-filter')."
                             (push (cons "formatted_body" formatted-body) it)
                             (push (cons "format" "org.matrix.custom.html") it))))
                (then (or then #'ignore)))
-    (when replying-to-event
-      (setf content (ement--add-reply content replying-to-event room)))
     (when filter
       (setf content (funcall filter content room)))
+    (when replying-to-event
+      (setf content (ement--add-reply content replying-to-event room)))
     (ement-api session endpoint :method 'put :data (json-encode content)
       :then (apply-partially then :room room :session session
                              ;; Data is added when calling back.
                              :content content :data))))
 
 (defalias 'ement--button-buttonize
+  ;; FIXME: This doesn't set the mouse-face to highlight, and it doesn't use the
+  ;; default-button category.  Neither does `button-buttonize', of course, but why?
   (if (version< emacs-version "28.1")
       (lambda (string callback &optional data)
         "Make STRING into a button and return it.
@@ -1262,13 +1351,13 @@ IMAGE should be one as created by, e.g. `create-image'."
 Works in major-modes `ement-room-mode',
 `ement-tabulated-room-list-mode', and `ement-room-list-mode'."
   (pcase major-mode
-    ('ement-room-mode (ement--format-room ement-room 'topic))
-    ('ement-tabulated-room-list-mode (ement--format-room (tabulated-list-get-id) 'topic))
+    ('ement-room-mode ement-room)
+    ('ement-tabulated-room-list-mode (tabulated-list-get-id))
     ('ement-room-list-mode
      (cl-typecase (oref (magit-current-section) value)
        (taxy-magit-section nil)
        (t (pcase (oref (magit-current-section) value)
-            (`[,room ,_session] (ement--format-room room 'topic))))))))
+            (`[,room ,_session] room)))))))
 
 (defun ement--room-direct-p (room session)
   "Return non-nil if ROOM on SESSION is a direct chat."
@@ -1339,6 +1428,11 @@ Works in major-modes `ement-room-mode',
                (heroes joined)
                (format "%s, and %s others" (hero-names heroes)
                        (- joined (length heroes))))
+              (name-override
+               () (when-let ((event (alist-get "org.matrix.msc3015.m.room.name.override"
+                                               (ement-room-account-data room)
+                                               nil nil #'equal)))
+                    (map-nested-elt event '(content name))))
               (empty-room
                (heroes joined) (cl-etypecase (length heroes)
                                  ((satisfies zerop) "Empty room")
@@ -1346,7 +1440,8 @@ Works in major-modes `ement-room-mode',
                                                        (hero-names heroes)))
                                  (t (format "Empty room (was %s)"
                                             (heroes-and-others heroes joined))))))
-    (or (latest-event "m.room.name" 'name)
+    (or (name-override)
+        (latest-event "m.room.name" 'name)
         (latest-event "m.room.canonical_alias" 'alias)
         (heroes-name)
         (member-events-name)
